"""Core operations on a CCHDO CF/netCDF file."""

from collections.abc import Hashable

import numpy as np
import numpy.typing as npt
import xarray as xr

from cchdo.hydro.checks import check_ancillary_variables
from cchdo.hydro.consts import (
    COORDS,
    CTDPRS,
    DIMS,
    FILLS_MAP,
)
from cchdo.hydro.dt import combine_dt
from cchdo.hydro.flags import (
    ExchangeBottleFlag,
    ExchangeCTDFlag,
    ExchangeFlag,
    ExchangeSampleFlag,
)
from cchdo.hydro.types import FileType
from cchdo.hydro.utils import (
    add_cdom_coordinate,
    add_geometry_var,
    add_profile_type,
    flatten_cdom_coordinate,
    set_axis_attrs,
    set_coordinate_encoding_fill,
)
from cchdo.params import WHPName, WHPNames

dtype_map = {"string": "U", "integer": "float32", "decimal": "float64"}

FLAG_SCHEME: dict[str, type[ExchangeFlag]] = {
    "woce_bottle": ExchangeBottleFlag,
    "woce_discrete": ExchangeSampleFlag,
    "woce_ctd": ExchangeCTDFlag,
}


def dataarray_factory(
    param: WHPName,
    ctype="data",
    N_PROF=0,
    N_LEVELS=0,
    strlen=0,
) -> xr.DataArray:
    dtype = dtype_map[param.dtype]
    if param.dtype == "string":
        dtype = f"U{strlen}"
    fill = FILLS_MAP[param.dtype]
    name = param.full_nc_name

    if ctype == "flag":
        dtype = dtype_map["integer"]
        fill = FILLS_MAP["integer"]
        name = param.nc_name_flag

    if param.scope == "profile":
        arr = np.full((N_PROF), fill_value=fill, dtype=dtype)
    if param.scope == "sample":
        arr = np.full((N_PROF, N_LEVELS), fill_value=fill, dtype=dtype)

    attrs = param.get_nc_attrs()
    if "C_format" in attrs:
        attrs["C_format_source"] = "database"

    if ctype == "error":
        attrs = param.get_nc_attrs(error=True)
        name = param.nc_name_error

    if ctype == "flag" and param.flag_w in FLAG_SCHEME:
        flag_defs = FLAG_SCHEME[param.flag_w]
        flag_values = []
        flag_meanings = []
        for flag in flag_defs:
            flag_values.append(int(flag))
            flag_meanings.append(flag.cf_def)

        odv_conventions_map = {
            "woce_bottle": "WOCESAMPLE - WOCE Quality Codes for the sampling device itself",
            "woce_ctd": "WOCECTD - WOCE Quality Codes for CTD instrument measurements",
            "woce_discrete": "WOCEBOTTLE - WOCE Quality Codes for water sample (bottle) measurements",
        }

        attrs = {
            "standard_name": "status_flag",
            "flag_values": np.array(flag_values, dtype="int8"),
            "flag_meanings": " ".join(flag_meanings),
            "conventions": odv_conventions_map[param.flag_w],
        }

    var_da = xr.DataArray(arr, dims=DIMS[: arr.ndim], attrs=attrs, name=name)

    if param.dtype != "decimal":
        try:
            del var_da.attrs["C_format"]
        except KeyError:
            pass
        try:
            del var_da.attrs["C_format_source"]
        except KeyError:
            pass

    if param.dtype == "string":
        var_da.encoding["dtype"] = "S1"

    if param.dtype == "integer":
        var_da.encoding["dtype"] = "int32"
        var_da.encoding["_FillValue"] = -999  # classic

    if param in COORDS and param != CTDPRS:
        var_da.encoding["_FillValue"] = None
        if param.dtype == "integer":
            var_da = var_da.fillna(-999).astype("int32")

    if ctype == "flag":
        var_da.encoding["dtype"] = "int8"
        var_da.encoding["_FillValue"] = 9

    var_da.encoding["zlib"] = True

    return var_da


def remove_param(
    ds: xr.Dataset,
    param: WHPName | str,
    *,
    delete_param: bool = False,
    delete_flag: bool = False,
    delete_error: bool = False,
    delete_ancillary=False,
    require_empty: bool = True,
) -> xr.Dataset:
    """Remove a parameter and/or its ancillary variables from a dataset.

    Since this function is rather destructive, by default, it is basically a no-op, you must make a choice when calling to delete the parameter or ancillary variable.

    :param ds: The dataset to remove a param from.
    :param param: A string or ``WHPName`` instance of a parameter, strings are in the form ``"param [unit]"`` or ``"param"`` if unitless, e.g. ``"OXYGEN [UMOL/KG]"`` or ``"EXPOCODE"``
    :param delete_param: If True, delete this parameter and all associated ancillary variables. Defaults to False.
    :param delete_flag: If True, delete this parameters flag, Defaults to False.
    :param delete_error: If True, delete this parameters error variable. Defaults to False.
    :param delete_ancillary: Currently a no-op.
    :param require_empty: If True, require that all the variables that will be deleted contain exclusively fill values, defualts to True

    :returns: A new dataset with the requested variables removed, will return a new dataset even if the results of calling this function are no-op
    """
    if isinstance(param, str):
        _param = WHPNames[param]
    else:
        _param = param

    # deleting the primary var requries the delete cascade
    if delete_param is True:
        delete_flag = True
        delete_error = True
        # TODO ancillary
        # delete_ancillary = True

    nc_name = _param.full_nc_name
    nc_flag = _param.nc_name_flag
    nc_error = _param.nc_name_error

    to_delete = []
    if delete_param:
        to_delete.append(nc_name)
    if delete_flag:
        to_delete.append(nc_flag)
    if delete_error:
        to_delete.append(nc_error)

    if require_empty:
        for _name in to_delete:
            if _name not in ds:
                continue
            data = ds[_name]
            # TODO: handle string arrays
            if not np.isnan(data).all():
                raise ValueError(f"{_name} is not empty")

    ds = ds.copy()
    ds = flatten_cdom_coordinate(ds)
    # TODO: this relies on the param name, should rely on attrs somehow
    base_param = ds[_param.full_nc_name]
    ancillary_vars = {
        name: ds[name]
        for name in base_param.attrs.get("ancillary_variables", "").split()
    }

    for _name in to_delete:
        try:
            del ds[_name]
        except KeyError:
            pass
        try:
            del ancillary_vars[_name]
        except KeyError:
            pass

    if nc_name in ds:
        new_ancillary_variables = " ".join(sorted(ancillary_vars))
        if new_ancillary_variables != "":
            ds[nc_name].attrs["ancillary_variables"] = new_ancillary_variables
        else:
            try:
                del ds[nc_name].attrs["ancillary_variables"]
            except KeyError:
                pass

    ds = add_cdom_coordinate(ds)
    check_ancillary_variables(ds)
    return ds


def add_param(
    ds: xr.Dataset,
    param: WHPName | str,
    *,
    with_flag: bool = False,
    with_error: bool = False,
    with_ancillary=None,
) -> xr.Dataset:
    """Add a new parameter, and optionally some ancillary associated variables to a dataset.

    This function is idempotent and will not overwrite any existing data or parameters.
    To add a flag or error parameter to an existing parameter is done by setting the `with_flag` or `with_error` arguments to `True`.

    :param ds: Dataset to add a param to
    :param param: A string or ``WHPName`` instance of a parameter, strings are in the form ``"param [unit]"`` or ``"param"`` if unitless, e.g. ``"OXYGEN [UMOL/KG]"`` or ``"EXPOCODE"``
    :param with_flag: If True, also add a quality flag variable, defaults to False
    :param with_error: If True, also add an error variable if one is defined for this parent parameter, defaults to False
    :param with_ancillary: Currently a no-op, hopefully will be used to add ancillary variable for things like analytical temperature
    :return: A new dataset with the requested variables added, will return a new dataset even if the results of calling this function are no-op"""
    if isinstance(param, str):
        _param = WHPNames[param]
    else:
        _param = param

    _ds = ds.copy()
    _ds = flatten_cdom_coordinate(_ds)
    vars_to_add = []

    if _param.full_nc_name in _ds:
        var = _ds[_param.full_nc_name]
    else:
        var = dataarray_factory(
            _param, N_PROF=ds.sizes["N_PROF"], N_LEVELS=ds.sizes["N_LEVELS"]
        )
        vars_to_add.append(var)

    if with_flag and _param.nc_name_flag not in _ds:
        flag_var = dataarray_factory(
            _param,
            N_PROF=ds.sizes["N_PROF"],
            N_LEVELS=ds.sizes["N_LEVELS"],
            ctype="flag",
        )
        ancillary = var.attrs.get("ancillary_variables", "").split()
        if flag_var.name not in ancillary:
            ancillary.append(flag_var.name)
        var.attrs["ancillary_variables"] = " ".join(sorted(ancillary))
        vars_to_add.append(flag_var)

    if with_error and _param.full_error_name is None:
        raise ValueError(f"{_param} does not have a defined error/uncertainty name")

    if with_error and _param.nc_name_error not in _ds:
        error_var = dataarray_factory(
            _param,
            N_PROF=ds.sizes["N_PROF"],
            N_LEVELS=ds.sizes["N_LEVELS"],
            ctype="error",
        )
        ancillary = var.attrs.get("ancillary_variables", "").split()
        if error_var.name not in ancillary:
            ancillary.append(error_var.name)
        var.attrs["ancillary_variables"] = " ".join(sorted(ancillary))
        vars_to_add.append(error_var)

    for var in vars_to_add:
        _ds[var.name] = var

    _ds = add_cdom_coordinate(_ds)
    check_ancillary_variables(_ds)

    return _ds


def add_profile_level(ds: xr.Dataset, idx, levels) -> xr.Dataset:
    return ds


def add_level(ds: xr.Dataset, n_levels=1) -> xr.Dataset:
    """Expand the size of N_LEVELS

    This essentially expands the slots avaialble for data and does not add any actual valid levels to a profile.
    """
    return ds.pad({"N_LEVELS": (0, n_levels)})


def add_profile(
    ds: xr.Dataset,
    expocode: npt.ArrayLike,
    station: npt.ArrayLike,
    cast: npt.ArrayLike,
    time: npt.ArrayLike,
    latitude: npt.ArrayLike,
    longitude: npt.ArrayLike,
    profile_type: npt.ArrayLike,
) -> xr.Dataset:
    """Add one or more profiles to the dataset.

    Arguments maybe be scalar or 1d arrays, all 1d arrays must be of the same size.
    Scalars will be broadcast against any 1d arrays
    """
    old_sizes = ds.sizes

    ds = ds.copy()

    (
        expocode,
        station,
        cast,
        time,
        latitude,
        longitude,
        profile_type,
    ) = np.broadcast_arrays(
        np.atleast_1d(expocode), station, cast, time, latitude, longitude, profile_type
    )

    new_profs: dict[Hashable, npt.NDArray] = {
        "expocode": expocode,
        "station": station,
        "cast": cast,
        "time": time.astype("datetime64[ns]"),  # ensure ns precision for now
        "latitude": latitude,
        "longitude": longitude,
        "profile_type": profile_type,
    }

    ds = ds.pad({"N_PROF": (0, len(expocode))})
    for param, value in new_profs.items():
        ds[param][old_sizes["N_PROF"] :] = value

    # TODO integrity tests?
    # TODO sorting profiles?
    return ds


def create_new() -> xr.Dataset:
    """Create an empty CF Dataset with the minimum required contents."""
    dataarrays = {}
    for param in COORDS:
        dataarrays[param.nc_name] = dataarray_factory(param)
    ds = xr.Dataset(dataarrays)

    ds = set_coordinate_encoding_fill(ds)

    ds = combine_dt(ds)

    ds = ds.set_coords([coord.nc_name for coord in COORDS if coord.nc_name in ds])

    ds = add_profile_type(ds, FileType.BOTTLE)  # just adds the var if no dims > 0
    ds = set_axis_attrs(ds)
    ds = add_geometry_var(ds)
    return ds



import io

import numpy as np
import pytest
import xarray as xr

from cchdo.hydro import core
from cchdo.hydro.exchange import read_exchange
from cchdo.hydro.exchange.helpers import simple_bottle_exchange
from cchdo.params import WHPNames


def test_create_new():
    expected = xr.Dataset(
        {
            "profile_type": xr.DataArray(
                np.empty((0), np.dtype("U")),
                dims=["N_PROF"],
                attrs={},
                name="profile_type",
            ),
            "geometry_container": xr.DataArray(
                np.array(np.nan, np.dtype("float64")),
                dims=[],
                attrs={
                    "geometry_type": "point",
                    "node_coordinates": "longitude latitude",
                },
            ),
        },
        coords={
            "expocode": xr.DataArray(
                np.empty((0), np.dtype("U")),
                dims=["N_PROF"],
                attrs={"whp_name": "EXPOCODE", "geometry": "geometry_container"},
            ),
            "station": xr.DataArray(
                np.empty((0), np.dtype("U")),
                dims=["N_PROF"],
                attrs={"whp_name": "STNNBR", "geometry": "geometry_container"},
            ),
            "cast": xr.DataArray(
                np.empty((0), np.dtype("int32")),
                dims=["N_PROF"],
                attrs={"whp_name": "CASTNO", "geometry": "geometry_container"},
            ),
            "sample": xr.DataArray(
                np.empty((0, 0), np.dtype("U")),
                dims=["N_PROF", "N_LEVELS"],
                attrs={"whp_name": "SAMPNO"},
            ),
            "time": xr.DataArray(
                np.empty((0), np.dtype("datetime64[ns]")),
                dims=["N_PROF"],
                attrs={
                    "standard_name": "time",
                    "whp_name": "DATE",
                    "resolution": 1,
                    "axis": "T",
                    "geometry": "geometry_container",
                },
            ),
            "latitude": xr.DataArray(
                np.empty((0), np.dtype("float64")),
                dims=["N_PROF"],
                attrs={
                    "standard_name": "latitude",
                    "units": "degree_north",
                    "C_format": "%9.4f",
                    "C_format_source": "database",
                    "axis": "Y",
                    "whp_name": "LATITUDE",
                },
            ),
            "longitude": xr.DataArray(
                np.empty((0), np.dtype("float64")),
                dims=["N_PROF"],
                attrs={
                    "standard_name": "longitude",
                    "units": "degree_east",
                    "C_format": "%9.4f",
                    "C_format_source": "database",
                    "axis": "X",
                    "whp_name": "LONGITUDE",
                },
            ),
            "pressure": xr.DataArray(
                np.empty((0, 0), np.dtype("float64")),
                dims=["N_PROF", "N_LEVELS"],
                attrs={
                    "standard_name": "sea_water_pressure",
                    "units": "dbar",
                    "C_format": "%9.1f",
                    "C_format_source": "database",
                    "axis": "Z",
                    "whp_name": "CTDPRS",
                    "whp_unit": "DBAR",
                    "positive": "down",
                },
            ),
        },
    )

    result = core.create_new()

    xr.testing.assert_identical(result, expected)


def test_add_profile():
    expected = xr.Dataset(
        {
            "profile_type": xr.DataArray(
                np.array(["B"], np.dtype("U")),
                dims=["N_PROF"],
                attrs={},
                name="profile_type",
            ),
            "geometry_container": xr.DataArray(
                np.array(np.nan, np.dtype("float64")),
                dims=[],
                attrs={
                    "geometry_type": "point",
                    "node_coordinates": "longitude latitude",
                },
            ),
        },
        coords={
            "expocode": xr.DataArray(
                np.array(["318M"], np.dtype("U")),
                dims=["N_PROF"],
                attrs={"whp_name": "EXPOCODE", "geometry": "geometry_container"},
            ),
            "station": xr.DataArray(
                np.array(["test"], np.dtype("U")),
                dims=["N_PROF"],
                attrs={"whp_name": "STNNBR", "geometry": "geometry_container"},
            ),
            "cast": xr.DataArray(
                np.array([1], np.dtype("int32")),
                dims=["N_PROF"],
                attrs={"whp_name": "CASTNO", "geometry": "geometry_container"},
            ),
            "sample": xr.DataArray(
                np.empty((1, 0), np.dtype("U")),
                dims=["N_PROF", "N_LEVELS"],
                attrs={"whp_name": "SAMPNO"},
            ),
            "time": xr.DataArray(
                np.array(["2020-01-01T00:00:00"], np.dtype("datetime64[ns]")),
                dims=["N_PROF"],
                attrs={
                    "standard_name": "time",
                    "whp_name": "DATE",
                    "resolution": 1,
                    "axis": "T",
                    "geometry": "geometry_container",
                },
            ),
            "latitude": xr.DataArray(
                np.array([0], np.dtype("float64")),
                dims=["N_PROF"],
                attrs={
                    "standard_name": "latitude",
                    "units": "degree_north",
                    "C_format": "%9.4f",
                    "C_format_source": "database",
                    "axis": "Y",
                    "whp_name": "LATITUDE",
                },
            ),
            "longitude": xr.DataArray(
                np.array([0], np.dtype("float64")),
                dims=["N_PROF"],
                attrs={
                    "standard_name": "longitude",
                    "units": "degree_east",
                    "C_format": "%9.4f",
                    "C_format_source": "database",
                    "axis": "X",
                    "whp_name": "LONGITUDE",
                },
            ),
            "pressure": xr.DataArray(
                np.empty((1, 0), np.dtype("float64")),
                dims=["N_PROF", "N_LEVELS"],
                attrs={
                    "standard_name": "sea_water_pressure",
                    "units": "dbar",
                    "C_format": "%9.1f",
                    "C_format_source": "database",
                    "axis": "Z",
                    "whp_name": "CTDPRS",
                    "whp_unit": "DBAR",
                    "positive": "down",
                },
            ),
        },
    )

    base = core.create_new()

    result = core.add_profile(base, "318M", "test", 1, "2020-01-01T00:00:00", 0, 0, "B")

    xr.testing.assert_identical(result, expected)


def test_add_param():
    # The easiest way I could think of to test this was to make some exchange inputs
    # one with and without some parameter and try use the functions to test the addition
    # and removal of a param
    # TODO: parameterize this to test the parameter space
    params = ("DELC14", "DELC14_FLAG_W", "C14ERR")
    units = ("/MILLE", "", "/MILLE")
    data = ("-999", "9", "-999")
    ds = read_exchange(
        io.BytesIO(simple_bottle_exchange()), precision_source="database"
    )
    ds_param = read_exchange(
        io.BytesIO(
            simple_bottle_exchange(params=params[:1], units=units[:1], data=data[:1])
        ),
        precision_source="database",
    )
    ds_param_flag = read_exchange(
        io.BytesIO(
            simple_bottle_exchange(params=params[:2], units=units[:2], data=data[:2])
        ),
        precision_source="database",
    )
    ds_param_flag_error = read_exchange(
        io.BytesIO(
            simple_bottle_exchange(params=params[:3], units=units[:3], data=data[:3])
        ),
        precision_source="database",
    )

    testing_ds_param = core.add_param(ds, WHPNames["DELC14 [/MILLE]"])
    xr.testing.assert_identical(ds_param, testing_ds_param)
    testing_ds_param_flag = core.add_param(
        ds, WHPNames["DELC14 [/MILLE]"], with_flag=True
    )
    xr.testing.assert_identical(ds_param_flag, testing_ds_param_flag)
    testing_ds_param_flag_error = core.add_param(
        ds, WHPNames["DELC14 [/MILLE]"], with_flag=True, with_error=True
    )
    xr.testing.assert_identical(ds_param_flag_error, testing_ds_param_flag_error)


@pytest.mark.parametrize("require_empty", [True, False])
@pytest.mark.parametrize("empty", [True, False])
@pytest.mark.parametrize("error", [True, False])
@pytest.mark.parametrize("flag", [True, False])
@pytest.mark.parametrize("param", [True, False])
def test_remove_param(param, flag, error, empty, require_empty):
    # TODO see above todo

    if param is True:
        flag, error = True, True

    bools = (param, flag, error)
    params = ("DELC14", "DELC14_FLAG_W", "C14ERR")
    units = ("/MILLE", "", "/MILLE")
    if empty:
        data = ("-999", "9", "-999")
    else:
        data = ("0.1", "2", "0.1")

    if not all(bools):
        _, _params, _units, _data = zip(
            *filter(lambda x: not x[0], zip(bools, params, units, data, strict=True)),
            strict=True,
        )
    else:
        _params = ()
        _units = ()
        _data = ()

    expected_ds = read_exchange(
        io.BytesIO(simple_bottle_exchange(params=_params, units=_units, data=_data)),
        precision_source="database",
    )
    input_ds = read_exchange(
        io.BytesIO(simple_bottle_exchange(params=params, units=units, data=data)),
        precision_source="database",
    )

    if empty is False and require_empty is True and any(bools):
        with pytest.raises(ValueError):
            testing_ds = core.remove_param(
                input_ds,
                "DELC14 [/MILLE]",
                delete_param=param,
                delete_flag=flag,
                delete_error=error,
                require_empty=require_empty,
            )
    else:
        testing_ds = core.remove_param(
            input_ds,
            "DELC14 [/MILLE]",
            delete_param=param,
            delete_flag=flag,
            delete_error=error,
            require_empty=require_empty,
        )
        xr.testing.assert_identical(expected_ds, testing_ds)


def test_add_param_cdom():
    params = (
        "CDOM300",
        "CDOM325",
    )
    units = (
        "/METER",
        "/METER",
    )
    data = (
        "-999",
        "-999",
    )
    ds = read_exchange(
        io.BytesIO(simple_bottle_exchange()), precision_source="database"
    )
    ds_param_300 = read_exchange(
        io.BytesIO(
            simple_bottle_exchange(params=params[:1], units=units[:1], data=data[:1])
        ),
        precision_source="database",
    )
    ds_param_325 = read_exchange(
        io.BytesIO(
            simple_bottle_exchange(params=params[1:], units=units[1:], data=data[1:])
        ),
        precision_source="database",
    )
    ds_param_cdom = read_exchange(
        io.BytesIO(simple_bottle_exchange(params=params, units=units, data=data)),
        precision_source="database",
    )

    testing_ds_param = core.add_param(ds, WHPNames["CDOM300 [/METER]"])
    xr.testing.assert_identical(ds_param_300, testing_ds_param)

    testing_ds_param = core.add_param(ds_param_300, WHPNames["CDOM325 [/METER]"])
    xr.testing.assert_identical(ds_param_cdom, testing_ds_param)

    testing_ds_param = core.add_param(ds, WHPNames["CDOM325 [/METER]"])
    xr.testing.assert_identical(ds_param_325, testing_ds_param)

    testing_ds_param = core.add_param(ds_param_325, WHPNames["CDOM300 [/METER]"])
    xr.testing.assert_identical(ds_param_cdom, testing_ds_param)



import pytest

from cchdo.hydro.checks import check_ancillary_variables


def test_check_ancillary_variables_ok(nc_placeholder):
    # passes if it doesn't raise
    check_ancillary_variables(nc_placeholder)


def test_check_ancillary_variables_unreferenced_vars(nc_placeholder):
    with pytest.raises(ValueError):
        data = nc_placeholder.copy()
        del data.del_carbon_14_dic.attrs["ancillary_variables"]
        check_ancillary_variables(data)


def test_check_ancillary_variables_referenced_non_extant_var(nc_placeholder):
    with pytest.raises(ValueError):
        data = nc_placeholder.copy()
        del data["del_carbon_14_dic_qc"]
        check_ancillary_variables(data)



from typing import Any, TypeGuard

import numpy as np
import numpy.typing as npt
import xarray as xr

from cchdo.hydro.consts import COORDS, DIMS, GEOMETRY_VARS, TIME
from cchdo.hydro.types import FileType, FileTypeType
from cchdo.params import WHPNames


def _is_all_dataarray(val: list[Any]) -> TypeGuard[list[xr.DataArray]]:
    return all(isinstance(obj, xr.DataArray) for obj in val)


def all_same(ndarr: np.ndarray) -> np.bool_:
    """Test if all the values of an ndarray are the same value"""
    if np.issubdtype(ndarr.dtype, np.number) and np.isnan(ndarr.flat[0]):
        return np.all(np.isnan(ndarr))
    return np.all(ndarr == ndarr.flat[0])


def add_profile_type(dataset: xr.Dataset, ftype: FileTypeType) -> xr.Dataset:
    """Adds a `profile_type` string variable to the dataset.

    This is for ODV compatability

    .. warning::
      Currently mixed profile types are not supported
    """
    ftype = FileType(ftype)

    profile_type = xr.DataArray(
        np.full(dataset.sizes["N_PROF"], fill_value=ftype.value, dtype="U1"),
        name="profile_type",
        dims=DIMS[0],
    )
    profile_type.encoding["dtype"] = "S1"

    dataset["profile_type"] = profile_type
    return dataset


def add_geometry_var(dataset: xr.Dataset) -> xr.Dataset:
    """Adds a CF-1.8 Geometry container variable to the dataset

    This allows for compatabiltiy with tools like gdal
    """
    dataset = dataset.copy()

    geometry_var = xr.DataArray(
        name="geometry_container",
        attrs={
            "geometry_type": "point",
            "node_coordinates": "longitude latitude",
        },
    )
    dataset["geometry_container"] = geometry_var

    for var in GEOMETRY_VARS:
        if var in dataset:
            dataset[var].attrs["geometry"] = "geometry_container"

    return dataset


def add_cdom_coordinate(dataset: xr.Dataset) -> xr.Dataset:
    """Find all the paraters in the cdom group and add their wavelength in a new coordinate"""

    cdom_names = list(filter(lambda x: x.nc_group == "cdom", WHPNames.values()))

    cdom_names = sorted(cdom_names, key=lambda x: x.radiation_wavelength or 0)

    # NM this needs to be sorted by wavelength...
    cdom_data = [
        dataset[name.full_nc_name]
        for name in cdom_names
        if name.full_nc_name in dataset
    ]

    # nothing to do
    if len(cdom_data) == 0:
        return dataset

    cdom_qc = [
        dataset.get(dataarray.attrs.get("ancillary_variables"))
        for dataarray in cdom_data
    ]

    # useful for later coping of attrs
    first = cdom_data[0]

    # "None in" doesn't seem to work due to xarray comparison?
    none_in_qc = [da is None for da in cdom_qc]
    if any(none_in_qc) and not all(none_in_qc):
        raise NotImplementedError("partial QC for CDOM is not handled yet")

    radiation_wavelengths = []
    c_formats = []
    for dataarray in cdom_data:
        whp_name = dataarray.attrs["whp_name"]
        whp_unit = dataarray.attrs["whp_unit"]
        whpname = WHPNames[(whp_name, whp_unit)]
        radiation_wavelengths.append(whpname.radiation_wavelength)
        if "C_format" in dataarray.attrs:
            c_formats.append(dataarray.attrs["C_format"])

    cdom_wavelengths = xr.DataArray(
        np.array(radiation_wavelengths, dtype=np.int32),
        dims="CDOM_WAVELENGTHS",
        name="CDOM_WAVELENGTHS",
        attrs={
            "standard_name": "radiation_wavelength",
            "units": "nm",
        },
    )

    new_cdom_dims = ("N_PROF", "N_LEVELS", "CDOM_WAVELENGTHS")
    new_cdom_coords = {
        "CDOM_WAVELENGTHS": cdom_wavelengths,
    }

    has_qc = False
    # qc flags first if any
    if _is_all_dataarray(cdom_qc):
        has_qc = True
        cdom_qc_arrays = np.stack(cdom_qc, axis=-1)
        first_qc = cdom_qc[0]

        new_cdom_qc_attrs = {**first_qc.attrs}
        new_cdom_qc = xr.DataArray(
            cdom_qc_arrays,
            dims=new_cdom_dims,
            coords=new_cdom_coords,
            attrs=new_cdom_qc_attrs,
        )
        new_cdom_qc.encoding = first_qc.encoding

    cdom_arrays = np.stack(cdom_data, axis=-1)

    new_cdom_attrs = {
        **first.attrs,
        "whp_name": "CDOM{CDOM_WAVELENGTHS}",
        "C_format": max(c_formats),
    }

    new_qc_name = f"{whpname.nc_group}_qc"

    if has_qc:
        new_cdom_attrs["ancillary_variables"] = new_qc_name

    new_cdom = xr.DataArray(
        cdom_arrays, dims=new_cdom_dims, coords=new_cdom_coords, attrs=new_cdom_attrs
    )
    new_cdom.encoding = first.encoding

    dataset[first.name] = new_cdom
    dataset = dataset.rename({first.name: whpname.nc_group})

    if has_qc:
        dataset[first_qc.name] = new_cdom_qc
        dataset = dataset.rename({first_qc.name: new_qc_name})

    for old_name in cdom_names:
        try:
            del dataset[old_name.full_nc_name]
        except KeyError:
            pass

    for old_name in cdom_names:
        try:
            del dataset[f"{old_name.full_nc_name}_qc"]
        except KeyError:
            pass

    return dataset


def flatten_cdom_coordinate(dataset: xr.Dataset) -> xr.Dataset:
    """Takes the a dataset with a CDOM wavelength and explocdes it back into individual variables"""
    if "cdom" not in dataset:
        return dataset

    keys = ["cdom", "CDOM_WAVELENGTHS"]
    if "cdom_qc" in dataset:
        keys.append("cdom_qc")

    ds = dataset.copy()
    cdom_var = ds[keys]
    for cdom_wavelength, arr in cdom_var.groupby("CDOM_WAVELENGTHS", squeeze=False):
        arr = arr.squeeze("CDOM_WAVELENGTHS")
        cdom = arr["cdom"].copy()

        cdom_qc = arr.get("cdom_qc")
        if cdom_qc is not None:
            cdom_qc = cdom_qc.copy()

        cdom.attrs["whp_name"] = cdom.attrs["whp_name"].format(
            CDOM_WAVELENGTHS=cdom_wavelength
        )

        whp_name = WHPNames[f"{cdom.attrs['whp_name']} [{cdom.attrs['whp_unit']}]"]

        if cdom_qc is not None:
            cdom.attrs["ancillary_variables"] = whp_name.nc_name_flag

        ds[whp_name.nc_name] = cdom
        if cdom_qc is not None:
            ds[whp_name.nc_name_flag] = cdom_qc

    return ds.drop_vars(keys)


def extract_numeric_precisions(
    data: list[str] | npt.NDArray[np.str_],
) -> npt.NDArray[np.int_]:
    """Get the numeric precision of a printed decimal number"""
    # magic number explain: np.char.partition expands each element into a 3-tuple
    # of (pre, sep, post) of some sep, in our case a "." char.
    # We only want the post bits [idx 2] (the number of chars after a decimal seperator)
    # of the last axis.
    numeric_parts = np.char.partition(data, ".")[..., 2]
    str_lens = np.char.str_len(numeric_parts)
    return np.max(str_lens, axis=0)


def set_axis_attrs(dataset: xr.Dataset) -> xr.Dataset:
    """Set the CF axis attribute on our axis variables (XYZT)

    * longitude = "X"
    * latitude = "Y"
    * pressure = "Z", addtionally, positive is down
    * time = "T"
    """
    dataset.longitude.attrs["axis"] = "X"
    dataset.latitude.attrs["axis"] = "Y"
    dataset.pressure.attrs["axis"] = "Z"
    dataset.pressure.attrs["positive"] = "down"
    dataset.time.attrs["axis"] = "T"
    return dataset


def set_coordinate_encoding_fill(dataset: xr.Dataset) -> xr.Dataset:
    """Sets the _FillValue encoidng to None for 1D coordinate vars"""
    for coord in COORDS:
        if coord is TIME and coord.nc_name not in dataset:
            continue
        if len(dataset[coord.nc_name].dims) == 1:
            dataset[coord.nc_name].encoding["_FillValue"] = None

    return dataset



from importlib.metadata import PackageNotFoundError, version

__version__: str = "999"

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    pass

from appdirs import AppDirs

_hydro_appdirs = AppDirs("edu.ucsd.cchdo.hydro")

from .exchange import read_csv, read_exchange

__all__ = ["read_csv", "read_exchange"]



import dataclasses
import io
import logging
from collections.abc import Iterable
from enum import Enum, auto
from functools import cached_property
from itertools import chain
from operator import attrgetter
from pathlib import Path
from typing import (
    TypedDict,
    TypeGuard,
)
from zipfile import ZipFile, is_zipfile

import numpy as np
import numpy.typing as npt
import requests
import xarray as xr

from cchdo.hydro.checks import check_flags
from cchdo.hydro.consts import (
    BTLNBR,
    CASTNO,
    COORDS,
    CTDPRS,
    EXPOCODE,
    FILLS_MAP,
    LATITUDE,
    SAMPNO,
    STNNBR,
    TIME,
)
from cchdo.hydro.core import dataarray_factory
from cchdo.hydro.dt import combine_dt
from cchdo.hydro.exchange.exceptions import (
    ExchangeBOMError,
    ExchangeDataInconsistentCoordinateError,
    ExchangeDataPartialCoordinateError,
    ExchangeDataPartialKeyError,
    ExchangeDuplicateKeyError,
    ExchangeDuplicateParameterError,
    ExchangeEncodingError,
    ExchangeFlaglessParameterError,
    ExchangeInconsistentMergeType,
    ExchangeMagicNumberError,
    ExchangeOrphanErrorError,
    ExchangeOrphanFlagError,
    ExchangeParameterUndefError,
    ExchangeParameterUnitAlignmentError,
)
from cchdo.hydro.flags import (
    ExchangeBottleFlag,
    ExchangeCTDFlag,
    ExchangeFlag,
    ExchangeSampleFlag,
)
from cchdo.hydro.sorting import sort_ds
from cchdo.hydro.types import (
    FileType,
    FileTypeType,
    PrecisionSource,
    PrecisionSourceType,
)
from cchdo.hydro.utils import (
    add_cdom_coordinate,
    add_geometry_var,
    add_profile_type,
    all_same,
    extract_numeric_precisions,
    set_axis_attrs,
    set_coordinate_encoding_fill,
)
from cchdo.params import WHPName, WHPNames
from cchdo.params import __version__ as params_version

try:
    from cchdo.hydro import __version__ as hydro_version

    CCHDO_VERSION = ".".join(hydro_version.split(".")[:2])
    if "dev" in hydro_version:
        CCHDO_VERSION = hydro_version
except ImportError:
    hydro_version = CCHDO_VERSION = "unknown"

log = logging.getLogger(__name__)


SENTINEL_PREFIX = "__SENTINEL"
SENTINEL_PARAM = WHPName(
    SENTINEL_PREFIX, SENTINEL_PREFIX, SENTINEL_PREFIX, -1, "string", False, 0
)

FLAG_SCHEME: dict[str, type[ExchangeFlag]] = {
    "woce_bottle": ExchangeBottleFlag,
    "woce_discrete": ExchangeSampleFlag,
    "woce_ctd": ExchangeCTDFlag,
}


# WHPNameIndex represents a Name to Column index in an exchange file
WHPNameIndex = dict[WHPName, int]
# WHPParamUnit represents the paired up contents of the Parameter and Unit lines
# in an exchange file
WHPParamUnit = tuple[str, str | None]


def _has_no_nones(val: list[str | None]) -> TypeGuard[list[str]]:
    return None not in val


def _transform_whp_to_csv(params: list[str], units: list[str]) -> list[str]:
    slots: list[str | None] = [None for _ in range(len(params))]

    pairs = list(zip(params, units, strict=True))
    mutable_units = list(units)

    if len(set(pairs)) != len(pairs):
        # we will assume that this is due to flags, actual duplciate params will come later
        for index, (param, unit) in enumerate(pairs):
            next_idx = index + 1
            try:
                next_param, next_unit = pairs[next_idx]
            except IndexError:
                continue

            potential_flag = f"{param}_FLAG_W"
            if next_param == potential_flag:
                mutable_units[next_idx] = unit

    flags = {}
    # param pass
    for index, (param, unit) in enumerate(zip(params, mutable_units, strict=True)):
        if param.endswith("_FLAG_W"):
            flags[param] = index
        slots[index] = f"{param} [{unit}]"

    # flag pass
    for param, unit in zip(params, mutable_units, strict=True):
        if param.endswith("_FLAG_W"):
            continue
        if (flag := f"{param}_FLAG_W") in flags:
            slots[flags[flag]] = f"{param} [{unit}]_FLAG_W"

    if _has_no_nones(slots):
        return slots

    raise ValueError("something has gone wrong with parameters transform")


def _get_params(
    params_units: Iterable[str], ignore_params: Iterable[str] | None = None
) -> tuple[WHPNameIndex, WHPNameIndex, WHPNameIndex]:
    params: WHPNameIndex = {}
    flags: WHPNameIndex = {}
    errors: WHPNameIndex = {}

    duplicate_errors = []
    unknown_errors = []

    if ignore_params is None:
        ignore_params = ()

    for index, param in enumerate(params_units):
        if param.startswith("__SENTINEL") or param in ignore_params:
            continue

        try:
            whpname = WHPNames[param]
        except KeyError:
            unknown_errors.append(param)
            continue

        if whpname.error_col:
            errors[whpname] = index

        elif whpname.flag_col:
            flags[whpname] = index

        else:
            if whpname in params:
                duplicate_errors.append(param)

            params[whpname] = index

    if any(unknown_errors):
        raise ExchangeParameterUndefError(unknown_errors)

    if any(duplicate_errors):
        raise ExchangeDuplicateParameterError(
            f"The following params are duplicate: {duplicate_errors}"
        )

    if not (params.keys() >= flags.keys()):
        raise ValueError(f"Some flags not in params: {flags.keys() - params.keys()}")
    if not (params.keys() >= errors.keys()):
        raise ValueError(f"Some errors not in params: {errors.keys() - params.keys()}")

    for flag in flags:
        if flag.flag_w == "no_flags":
            raise ExchangeFlaglessParameterError(flag)

    return params, flags, errors


def _ctd_get_header(line, dtype=str):
    header, value = (part.strip() for part in line.split("="))
    if header in ("_SAMPLING_RATE", "SAMPLING_RATE") and value.lower().endswith("hz"):
        value = value.rstrip(" HZhz")
    return header, dtype(value)


def finalize_ancillary_variables(dataset: xr.Dataset):
    """Turn the ancillary variable attr into a space seperated string

    It is nice to have the ancillary variable be a list while things are being read into it
    """
    for var in dataset.variables:
        if "ancillary_variables" not in dataset[var].attrs:
            continue
        ancillary_variables = dataset[var].attrs["ancillary_variables"]
        if len(ancillary_variables) == 0:
            del dataset[var].attrs["ancillary_variables"]
        elif isinstance(ancillary_variables, str):
            pass
        elif isinstance(ancillary_variables, list):
            dataset[var].attrs["ancillary_variables"] = " ".join(
                sorted(set(ancillary_variables))
            )
        else:
            raise ValueError("ancillary variables are crazy")

    return dataset


def combine_bottle_time(dataset: xr.Dataset):
    """Combine the bottle dates and times if present

    Raises if only one is present
    """
    BTL_TIME = WHPNames["BTL_TIME"]
    BTL_DATE = WHPNames["BTL_DATE"]

    if BTL_DATE.full_nc_name not in dataset and BTL_TIME.full_nc_name not in dataset:
        return dataset

    if BTL_TIME.nc_name in dataset and BTL_DATE.nc_name not in dataset:
        dates = np.char.replace(
            np.datetime_as_string(dataset[TIME.nc_name].values, unit="D"), "-", ""
        )

        dataset[BTL_DATE.nc_name] = dataset[BTL_TIME.nc_name].copy().astype("U8")
        dataset[BTL_DATE.nc_name].values.T[:] = dates
        dataset[BTL_DATE.nc_name].values[dataset[BTL_TIME.nc_name].values == ""] = ""

    ds = combine_dt(
        dataset,
        is_coord=False,
        date_name=BTL_DATE,
        time_name=BTL_TIME,
        time_pad=True,
    )

    # Take the station time as BO and go back one hour for "safty"
    reference_time = ds.time - np.timedelta64(1, "h")

    # Add a day to anything before the ref time
    next_day = (ds.bottle_time < reference_time).values
    ds.bottle_time.values[next_day] = ds.bottle_time.values[next_day] + np.timedelta64(
        1, "D"
    )

    return ds


def check_is_subset_shape(
    a1: npt.NDArray, a2: npt.NDArray, strict="disallowed"
) -> npt.NDArray[np.bool_]:
    """Ensure that the shape of the data in a2 is a subset (or strict subset) of the data shape of a1

    For a given set of param, flag, and error arrays you would want to ensure that:

    * errors are a subset of params (strict is allowed)
    * params are a subset of flags (strict is allowed)

    For string vars, the empty string is considered the "nothing" value.
    For woce flags, flag 9s should be converted to nans (depending on scheme flag 5 and 1 may not have param values)

    Return a boolean array of invalid locations
    """
    if a1.shape != a2.shape:
        raise ValueError("Cannot compare diffing shaped arrays")

    a1_values = np.isfinite(a1)
    a2_values = np.isfinite(a2)

    return a1_values != a2_values


@dataclasses.dataclass
class _ExchangeData:
    """Dataclass containing exchange data which has been parsed into ndarrays"""

    single_profile: bool
    param_cols: dict[WHPName, np.ndarray]
    flag_cols: dict[WHPName, np.ndarray]
    error_cols: dict[WHPName, np.ndarray]

    # OG Print Precition Tracking
    param_precisions: dict[WHPName, npt.NDArray[np.int_]]
    error_precisions: dict[WHPName, npt.NDArray[np.int_]]

    comments: str

    def __post_init__(self):
        # check the shapes of all the nd arrays are the same
        get_shape = attrgetter("shape")
        shapes = [
            get_shape(arr)
            for arr in chain(
                self.param_cols.values(),
                self.flag_cols.values(),
                self.error_cols.values(),
            )
        ]
        if not all(shape == shapes[0] for shape in shapes):
            # TODO Error handling
            raise ValueError("shape error")

        self.shape = shapes[0]

        if self.single_profile:
            # all "profile scoped" params must have the same values
            for param, data in self.param_cols.items():
                if param.scope != "profile":
                    continue
                if not np.unique(data).shape[0] == 1:
                    raise ValueError(f"inconsistent {param} {data}")

            # sample must be unique
            try:
                sample_ids = self.param_cols[SAMPNO]
            except KeyError as err:
                log.debug("SAMPNO not in file, attempting BTLNBR fallback")
                if BTLNBR in self.param_cols:
                    sample_ids = self.param_cols[BTLNBR]
                    self.param_cols[SAMPNO] = self.param_cols[BTLNBR]
                else:
                    raise ExchangeDataPartialKeyError("Missing SAMPNO") from err

            unique_sample_ids, unique_sample_counts = np.unique(
                sample_ids, return_counts=True
            )
            if unique_sample_ids.shape != sample_ids.shape:
                duplicated_values = unique_sample_ids[unique_sample_counts > 1]
                raise ExchangeDuplicateKeyError(
                    {
                        "EXPOCODE": self.param_cols[EXPOCODE][0],
                        "STNNBR": self.param_cols[STNNBR][0],
                        "CASTNO": self.param_cols[CASTNO][0],
                        "SAMPNO": str(duplicated_values),
                    }
                )

            # check coordinates are "full"
            for coord in COORDS:
                if coord is TIME and TIME not in self.param_cols:
                    continue
                data = self.param_cols[coord]
                if data.dtype.char in {"S", "U"}:
                    if np.any(data == ""):
                        raise ExchangeDataPartialCoordinateError(
                            f"{coord} has missing values"
                        )
                elif np.any(np.isnan(data)):
                    raise ExchangeDataPartialCoordinateError(
                        f"{coord} has missing values"
                    )

        # make sure flags and errors are strict subsets
        if not self.flag_cols.keys() <= self.param_cols.keys():
            raise ExchangeOrphanFlagError()
        if not self.error_cols.keys() <= self.param_cols.keys():
            raise ExchangeOrphanErrorError()

    def set_expected(
        self, params: set[WHPName], flags: set[WHPName], errors: set[WHPName]
    ):
        """Puts fill columns for expected params which are missing

        This can occur when there are disjoint columns in CTD files
        """
        ref_cols = {
            "string": EXPOCODE,
            "integer": CASTNO,
            "decimal": LATITUDE,
        }

        # we need to detect if just the flag is misisng and set to flag 0 or 9 depending on where data are
        # else set to flag 9
        for name in flags:
            if name in self.flag_cols:
                continue
            self.flag_cols[name] = np.full_like(
                self.param_cols[ref_cols["integer"]], fill_value=np.nan
            )
            if name in self.param_cols:
                self.flag_cols[name][np.isfinite(self.param_cols[name])] = 0

        for name in params:
            if name in self.param_cols:
                continue
            self.param_cols[name] = np.full_like(
                self.param_cols[ref_cols[name.dtype]], fill_value=FILLS_MAP[name.dtype]
            )

        for name in errors:
            if name in self.error_cols:
                continue
            self.error_cols[name] = np.full_like(
                self.param_cols[ref_cols[name.dtype]], fill_value=FILLS_MAP[name.dtype]
            )

    def split_profiles(self):
        """Split into single profile containing _ExchangeData instances

        Done by looking at the expocode+station+cast composate keys
        """
        try:
            expocode = self.param_cols[EXPOCODE]
        except KeyError as err:
            raise ExchangeDataPartialKeyError("Missing EXPOCODE") from err
        try:
            station = self.param_cols[STNNBR]
        except KeyError as err:
            raise ExchangeDataPartialKeyError("Missing STNNBR") from err
        try:
            cast = self.param_cols[CASTNO]
        except KeyError as err:
            raise ExchangeDataPartialKeyError("Missing CASTNO") from err

        # need to split up by profiles and _not_ assume the bottles are in order
        # use the actual values to sort things out
        # we don't care what the values are, they just need to work
        log.debug("Grouping Profiles by Key")
        # we need to add seperators to avoid conflicts
        # TODO: add test for when these might conflict
        expocode_sep = np.char.add(expocode, ",")
        station_sep = np.char.add(station, ",")
        # numpy concat basically
        prof_ids = np.char.add(np.char.add(expocode_sep, station_sep), cast.astype("U"))
        unique_profile_ids = np.unique(prof_ids)
        log.debug("Found %s unique profile keys", len(unique_profile_ids))
        profiles = [np.nonzero(prof_ids == prof) for prof in unique_profile_ids]

        log.debug("Actually splitting profiles")
        return [
            _ExchangeData(
                single_profile=True,
                param_cols={
                    param: data[profile] for param, data in self.param_cols.items()
                },
                flag_cols={
                    param: data[profile] for param, data in self.flag_cols.items()
                },
                error_cols={
                    param: data[profile] for param, data in self.error_cols.items()
                },
                param_precisions=self.param_precisions,
                error_precisions=self.error_precisions,
                comments=self.comments,
            )
            for profile in profiles
        ]

    @cached_property
    def str_lens(self) -> dict[WHPName, int]:
        """Figure out the length of all the string params

        The char size can vary by platform.
        """
        lens = {}
        for param, data in self.param_cols.items():
            if param.dtype == "string":
                lens[param] = int(np.max(np.char.str_len(data)))

        return lens


def _get_fill_locs(arr, fill_values: tuple[str, ...] = ("-999",)):
    fill = np.char.startswith(arr, fill_values[0])
    if len(fill_values) > 1:
        for fill_value in fill_values[1:]:
            fill = fill | np.char.startswith(arr, fill_value)
    return fill


@dataclasses.dataclass
class _ExchangeInfo:
    """Low level dataclass containing the parts of an exchange file"""

    stamp_slice: slice
    comments_slice: slice
    ctd_headers_slice: slice
    params_idx: int
    units_idx: int
    data_slice: slice
    post_data_slice: slice
    _raw_lines: tuple[str, ...] = dataclasses.field(repr=False)
    _ctd_override: bool = False
    _ignore_columns: Iterable[str] = ()

    @property
    def stamp(self):
        """Returns the filestamp of the exchange file

        e.g. "BOTTLE,20210301CCHSIOAMB"
        """
        return self._raw_lines[self.stamp_slice]

    @property
    def comments(self):
        """Returns the comments of the exchange file with leading # stripped"""
        raw_comments = self._raw_lines[self.comments_slice]
        return [c.removeprefix("#") for c in raw_comments]

    @property
    def ctd_headers(self):
        """Returns a dict of the CTD headers and their value"""
        return dict(
            [_ctd_get_header(line) for line in self._raw_lines[self.ctd_headers_slice]]
        )

    @cached_property
    def params(self):
        """Returns a list of all parameters in the file (including CTD "headers")"""
        ctd_params = self.ctd_headers.keys()
        data_params = self._raw_lines[self.params_idx].split(",")
        return [param.strip() for param in [*ctd_params, *data_params]]

    @cached_property
    def units(self):
        """Returns a list of all the units in the file (including CTD "headers")

        Will have the same shape as params
        """
        if self.params_idx == self.units_idx:
            return self.params

        # we can have a bunch of empty strings as units, we want these to be
        # None to match what would be in a WHPName object
        ctd_units = [None for _ in self.ctd_headers]
        data_units = self._raw_lines[self.units_idx].split(",")
        return [
            x if x != "" else None
            for x in [
                *ctd_units,
                *[unit.strip() for unit in data_units],
            ]
        ]

    @property
    def data(self):
        """Returns the data block of an exchange file as a tuple of strs.
        One line per entry.
        """
        return self._raw_lines[self.data_slice]

    @property
    def post_data(self):
        """Returns any post data content as a tuple of strs"""
        return self._raw_lines[self.post_data_slice]

    @cached_property
    def _whp_param_info(self):
        """Parses the params and units for base parameters

        Returns a dict with a WHPName to column index mapping
        """
        # In initial testing, it was discovered that approx half the ctd files
        # had trailing commas in just the params and units lines
        if self.params[-1] == "" and self.units[-1] is None:
            self.params.pop()
            self.units.pop()

        # the number of expected columns is just going to be the number of
        # parameter names we see
        column_count = len(self.params)

        if len(self.units) != column_count:
            if len(self.units) > column_count:
                # attempt to fix trailing commas in units (assume PARAMS is canonical)
                while len(self.units) > column_count:
                    if self.units[-1] is not None:
                        break
                    self.units.pop()

            # check to see if above fixed it
            if len(self.units) != column_count:
                raise ExchangeParameterUnitAlignmentError

        if self.params_idx == self.units_idx:
            params_units = self.params
        else:
            params_units = _transform_whp_to_csv(self.params, self.units)
        params_idx, flags, errors = _get_params(params_units, self._ignore_columns)

        if any(self.ctd_headers) or self._ctd_override:
            params_idx[SAMPNO] = params_idx[CTDPRS]

        return params_idx, flags, errors

    @property
    def whp_params(self):
        return self._whp_param_info[0]

    @property
    def whp_flags(self):
        """Parses the params and units for flag values

        returns a dict with a WHPName to column index of flags mapping
        """
        return self._whp_param_info[1]

    @property
    def whp_errors(self):
        """Parses the params and units for uncertanty values

        returns a dict with a WHPName to column index of errors mapping
        """
        return self._whp_param_info[2]

    @property
    def _np_data_block(self):
        _raw_data = tuple(
            (*self.ctd_headers.values(), *line.replace(" ", "").split(","))
            for line in self.data
        )
        return np.array(_raw_data, dtype="U")

    def finalize(
        self,
        fill_values=("-999",),
        precision_source: PrecisionSourceType = PrecisionSource.FILE,
    ) -> _ExchangeData:
        """Parse all the data into ndarrays of the correct dtype and shape

        Returns an ExchangeData dataclass
        """
        precision_source = PrecisionSource(precision_source)

        log.debug("Finializing...")
        single_profile = any(self.ctd_headers)

        np_db = self._np_data_block

        dtype_map = {"string": "U", "integer": "float32", "decimal": "float64"}

        whp_param_cols = {}
        whp_flag_cols = {}
        whp_error_cols = {}
        whp_param_precisions = {}
        whp_error_precisions = {}

        for param, idx in self.whp_params.items():
            param_col = np_db[:, idx]
            fill_spaces = _get_fill_locs(param_col, fill_values)
            if param.dtype in ("decimal", "integer"):
                if not _is_valid_exchange_numeric(param_col):
                    raise ValueError(
                        f"exchange numeric data for {param.whp_name} has bad chars"
                    )
                if precision_source == PrecisionSource.FILE:
                    whp_param_precisions[param] = extract_numeric_precisions(param_col)
                param_col[fill_spaces] = "nan"
            if param.dtype == "string":
                param_col[fill_spaces] = ""
            whp_param_cols[param] = param_col.astype(dtype_map[param.dtype])

        for param, idx in self.whp_flags.items():
            param_col = np_db[:, idx]
            fill_spaces = np.char.startswith(param_col, "9")
            param_col[fill_spaces] = "nan"
            whp_flag_cols[param] = np_db[:, idx].astype("float16")

        for param, idx in self.whp_errors.items():
            param_col = np_db[:, idx]
            fill_spaces = _get_fill_locs(param_col, fill_values)
            if param.dtype in ("decimal", "integer"):
                if not _is_valid_exchange_numeric(param_col):
                    raise ValueError(
                        f"{param} error col exchange numeric data has bad chars"
                    )
                if precision_source == PrecisionSource.FILE:
                    whp_error_precisions[param] = extract_numeric_precisions(param_col)
                param_col[fill_spaces] = "nan"
            whp_error_cols[param] = param_col.astype(dtype_map[param.dtype])

        comments = "\n".join([*self.stamp, *self.comments])
        del self._raw_lines

        return _ExchangeData(
            single_profile,
            whp_param_cols,
            whp_flag_cols,
            whp_error_cols,
            whp_param_precisions,
            whp_error_precisions,
            comments=comments,
        )

    @classmethod
    def from_lines(cls, lines: tuple[str, ...], ftype: FileTypeType, ignore_columns):
        """Figure out the line numbers/indicies of the parts of the exchange file"""
        stamp = 0  # file stamp is always the first line of a valid exchange
        comments_start = 1
        comments_end = 1
        ctd_header_start = 1
        ctd_header_end = 1
        params = 1
        units = 1
        data_start = 1
        data_end = 1
        post_data_start = 1
        post_data_end = 1

        ftype = FileType(ftype)

        class LookingFor(Enum):
            """States for the FSM that is this parser"""

            FILE_STAMP = auto()
            COMMENTS = auto()
            CTD_HEADERS = auto()
            PARAMS = auto()
            UNITS = auto()
            DATA = auto()
            POST_DATA = auto()

        state = LookingFor.FILE_STAMP
        ctd_num_headers = 0

        log.debug("Looking for file parts")

        for idx, line in enumerate(lines):
            if state is LookingFor.FILE_STAMP:
                state = LookingFor.COMMENTS
                continue

            if state is LookingFor.COMMENTS:
                if line.startswith("#"):
                    comments_end = idx + 1
                elif ftype == FileType.CTD:
                    state = LookingFor.CTD_HEADERS
                    param, value = _ctd_get_header(line, dtype=int)
                    if param != "NUMBER_HEADERS":
                        raise ValueError()
                    ctd_num_headers = value - 1
                    ctd_header_start = idx + 1
                    continue
                else:
                    state = LookingFor.PARAMS
                    continue

            if state is LookingFor.CTD_HEADERS:
                if ctd_num_headers == 0:
                    ctd_header_end = idx
                    state = LookingFor.PARAMS
                    continue
                ctd_num_headers -= 1

            if state is LookingFor.PARAMS:
                params = idx - 1
                state = LookingFor.UNITS
                continue

            if state is LookingFor.UNITS:
                units = idx - 1
                data_start = idx
                state = LookingFor.DATA
                continue

            if state is LookingFor.DATA:
                if line == "END_DATA":
                    data_end = idx

                    state = LookingFor.POST_DATA
                    post_data_start = post_data_end = idx + 1
                    continue

            if state is LookingFor.POST_DATA:
                post_data_end = idx

        return cls(
            stamp_slice=slice(stamp, comments_start),
            comments_slice=slice(comments_start, comments_end),
            ctd_headers_slice=slice(ctd_header_start, ctd_header_end),
            params_idx=params,
            units_idx=units,
            data_slice=slice(data_start, data_end),
            post_data_slice=slice(post_data_start, post_data_end),
            _raw_lines=lines,
            _ignore_columns=ignore_columns,
        )


def _is_valid_exchange_numeric(data: npt.NDArray[np.str_]) -> np.bool_:
    # see allowed code points of the exchange doc
    # essentially, only %f types (not %g)
    allowed_exchange_numeric_data_chars = [
        c.encode("utf-8") for c in list("0123456789.-")
    ] + [b""]
    aligned = np.require(data, requirements=["C_CONTIGUOUS"])
    return np.all(np.isin(aligned.view("|S1"), allowed_exchange_numeric_data_chars))


ExchangeIO = str | Path | io.BufferedIOBase


def _load_raw_exchange(
    filename_or_obj: ExchangeIO,
    *,
    file_seperator: str | None = None,
    keep_seperator=True,
    encoding="utf8",
) -> list[str]:
    if isinstance(filename_or_obj, str) and filename_or_obj.startswith("http"):
        log.info("Loading object over http")
        data_raw = io.BytesIO(requests.get(filename_or_obj).content)

    elif isinstance(filename_or_obj, str | Path) and Path(filename_or_obj).exists():
        log.info("Loading object from local file path")
        with open(filename_or_obj, "rb") as local_file:
            data_raw = io.BytesIO(local_file.read())

    # lets just try "reading"
    elif hasattr(filename_or_obj, "read"):
        log.info("Loading object open file object")
        data_raw = io.BytesIO(filename_or_obj.read())

    elif isinstance(filename_or_obj, bytes | bytearray):
        log.info("Loading raw data bytes")
        data_raw = io.BytesIO(filename_or_obj)

    data: list[str] = []

    if file_seperator is not None:
        data = data_raw.read().decode(encoding).strip().split(file_seperator)
        data = list(filter(lambda x: x != "", data))

        if keep_seperator:
            data = [(datum + file_seperator).strip() for datum in data]

        data = [datum.strip() for datum in data]

    elif is_zipfile(data_raw):
        data_raw.seek(0)  # is_zipfile moves the "tell" position
        with ZipFile(data_raw) as zipfile:
            for zipinfo in zipfile.infolist():
                log.debug("Reading %s", zipinfo)
                try:
                    data.append(zipfile.read(zipinfo).decode(encoding))
                except UnicodeDecodeError as error:
                    raise ExchangeEncodingError from error
    else:
        data_raw.seek(0)  # is_zipfile moves the "tell" position
        try:
            data.append(data_raw.read().decode(encoding))
        except UnicodeDecodeError as error:
            raise ExchangeEncodingError from error

    # cleanup the data_raw to free the memory
    data_raw.close()
    return data


class CheckOptions(TypedDict, total=False):
    """Flags and config that controll how strict the file checks are"""

    flags: bool


def read_csv(
    filename_or_obj: ExchangeIO,
    *,
    fill_values=("-999",),
    ftype: FileTypeType = FileType.BOTTLE,
    checks: CheckOptions | None = None,
    precision_source: PrecisionSourceType = PrecisionSource.FILE,
    encoding="utf8",
    ignore_columns: Iterable[str] | None = None,
) -> xr.Dataset:
    precision_source = PrecisionSource(precision_source)
    ftype = FileType(ftype)

    _checks: CheckOptions = {"flags": True}
    if checks is not None:
        _checks.update(checks)

    if ignore_columns is None:
        ignore_columns = ()

    data = _load_raw_exchange(
        filename_or_obj,
        file_seperator="something_very_unliekly~~~",
        keep_seperator=False,
        encoding=encoding,
    )

    if len(data) != 1:
        raise ValueError("read_csv can only read a single file")

    splitdata = data[0].splitlines()

    params_units = splitdata[0]
    whp_params: list[WHPName] = []

    sentinel_count = 0
    for name in params_units.split(","):
        if name in ignore_columns:
            whp_params.append(SENTINEL_PARAM.as_depth(sentinel_count + 1))
            sentinel_count += 1
        else:
            whp_params.append(WHPNames[name])

    params_units_list = [name.odv_key for name in whp_params]

    NONE_SLICE = slice(
        0,
        0,
    )
    new_data = (",".join(params_units_list), *splitdata[1:])
    exchange_data = _ExchangeInfo(
        stamp_slice=NONE_SLICE,
        comments_slice=NONE_SLICE,
        ctd_headers_slice=NONE_SLICE,
        params_idx=0,
        units_idx=0,
        data_slice=slice(1, None),
        post_data_slice=NONE_SLICE,
        _raw_lines=new_data,
        _ctd_override=ftype == FileType.CTD,
    ).finalize(precision_source=precision_source, fill_values=fill_values)
    return _from_exchange_data([exchange_data], ftype=ftype, checks=_checks)


def read_exchange(
    filename_or_obj: ExchangeIO,
    *,
    fill_values=("-999",),
    checks: CheckOptions | None = None,
    precision_source: PrecisionSourceType = PrecisionSource.FILE,
    file_seperator=None,
    keep_seperator=True,
    encoding="utf8",
    ignore_columns: Iterable[str] | None = None,
) -> xr.Dataset:
    """Loads the data from filename_or_obj and returns a xr.Dataset with the CCHDO
    CF/netCDF structure"""

    precision_source = PrecisionSource(precision_source)

    _checks: CheckOptions = {"flags": True}
    if checks is not None:
        _checks.update(checks)

    log.debug(f"Check options: {_checks}")

    data = _load_raw_exchange(
        filename_or_obj,
        file_seperator=file_seperator,
        keep_seperator=keep_seperator,
        encoding=encoding,
    )

    log.info("Checking for BOM")
    if any(df.startswith("\ufeff") for df in data):
        raise ExchangeBOMError

    log.info("Detecting file type")
    if all(df.startswith("BOTTLE") for df in data):
        ftype = FileType.BOTTLE
    elif all(df.startswith("CTD") for df in data):
        ftype = FileType.CTD
    elif all(df.startswith(("CTD", "BOTTLE")) for df in data):
        # Mixed CTD and BOTTLE files (probably in a zip)
        raise ExchangeInconsistentMergeType
    else:
        raise ExchangeMagicNumberError

    log.info("Found filetype: %s", ftype.name)

    exchange_data = [
        _ExchangeInfo.from_lines(
            tuple(df.splitlines()), ftype=ftype, ignore_columns=ignore_columns
        ).finalize(
            fill_values=fill_values,
            precision_source=precision_source,
        )
        for df in data
    ]

    return _from_exchange_data(exchange_data, ftype=ftype, checks=_checks)


def _from_exchange_data(
    exchange_data: list[_ExchangeData],
    *,
    ftype: FileTypeType = FileType.BOTTLE,
    checks: CheckOptions | None = None,
) -> xr.Dataset:
    _checks: CheckOptions = {"flags": True}
    if checks is not None:
        _checks.update(checks)

    if not all(fp.single_profile for fp in exchange_data):
        exchange_data = list(chain(*[exd.split_profiles() for exd in exchange_data]))

    N_PROF = len(exchange_data)
    N_LEVELS = max(fp.shape[0] for fp in exchange_data)

    log.debug((N_PROF, N_LEVELS))

    params = set(chain(*[exd.param_cols.keys() for exd in exchange_data]))
    flags = set(chain(*[exd.flag_cols.keys() for exd in exchange_data]))
    errors = set(chain(*[exd.error_cols.keys() for exd in exchange_data]))
    for exd in exchange_data:
        exd.set_expected(params, flags, errors)

    # to init the empty data arrays, we need to know the max size of all the string type parameters
    # otherwise the values will be silently truncated
    str_lens: dict[WHPName, int] = {}
    for exd in exchange_data:
        for param, length in exd.str_lens.items():
            if str_lens.get(param, 0) <= length:
                str_lens[param] = length
    log.debug(f"Total string lengths: {str_lens}")

    log.debug("Init DataArrays")
    dataarrays = {}
    for param in sorted(params):
        dataarrays[param.full_nc_name] = dataarray_factory(
            param, N_PROF=N_PROF, N_LEVELS=N_LEVELS, strlen=str_lens.get(param)
        )

        dataarrays[param.full_nc_name].attrs["ancillary_variables"] = []
        if param in flags:
            qc_name = param.nc_name_flag
            dataarrays[qc_name] = dataarray_factory(
                param, ctype="flag", N_PROF=N_PROF, N_LEVELS=N_LEVELS
            )
            dataarrays[param.full_nc_name].attrs["ancillary_variables"].append(qc_name)

        if param in errors:
            error_name = param.nc_name_error
            dataarrays[error_name] = dataarray_factory(
                param, ctype="error", N_PROF=N_PROF, N_LEVELS=N_LEVELS
            )
            dataarrays[param.full_nc_name].attrs["ancillary_variables"].append(
                error_name
            )

        # Check for ancillary temperature data and connect to the parent
        if param.analytical_temperature_name is not None:
            ancilary_temp_param = WHPNames[
                (param.analytical_temperature_name, param.analytical_temperature_units)
            ]
            if ancilary_temp_param in params:
                dataarrays[param.full_nc_name].attrs["ancillary_variables"].append(
                    ancilary_temp_param.full_nc_name
                )

    log.debug("Put data in arrays")
    comments = exchange_data[0].comments
    for n_prof, exd in enumerate(exchange_data):
        if exd.comments != comments:
            comments = f"{comments}\n----file_break----\n{exd.comments}"

        for param in params:
            if param in exd.param_precisions and param.dtype == "decimal":
                dataarrays[param.full_nc_name].attrs["C_format"] = (
                    f"%.{exd.param_precisions[param]}f"
                )
                dataarrays[param.full_nc_name].attrs["C_format_source"] = "input_file"
            if param in exd.error_precisions and param.dtype == "decimal":
                dataarrays[param.nc_name_error].attrs["C_format"] = (
                    f"%.{exd.error_precisions[param]}f"
                )
                dataarrays[param.nc_name_error].attrs["C_format_source"] = "input_file"

            if param.scope == "profile":
                if not all_same(exd.param_cols[param]):
                    raise ExchangeDataInconsistentCoordinateError(param)
                dataarrays[param.full_nc_name][n_prof] = exd.param_cols[param][0]

                if param in flags:
                    dataarrays[param.nc_name_flag][n_prof] = exd.flag_cols[param][0]
                if param in errors:
                    dataarrays[param.nc_name_error][n_prof] = exd.error_cols[param][0]

            if param.scope == "sample":
                data = exd.param_cols[param]
                dataarrays[param.full_nc_name][n_prof, : len(data)] = data

                if param in flags:
                    data = exd.flag_cols[param]
                    dataarrays[param.nc_name_flag][n_prof, : len(data)] = data
                if param in errors:
                    data = exd.error_cols[param]
                    dataarrays[param.nc_name_error][n_prof, : len(data)] = data

    ex_dataset = xr.Dataset(
        dataarrays,
        attrs={
            "Conventions": f"CF-1.8 CCHDO-{CCHDO_VERSION}",
            "cchdo_software_version": f"hydro {hydro_version}",
            "cchdo_parameters_version": f"params {params_version}",
            "comments": comments,
            "featureType": "profile",
        },
    )

    # The order of the following is somewhat important
    ex_dataset = set_coordinate_encoding_fill(ex_dataset)
    ex_dataset = combine_dt(ex_dataset, time_pad=True)

    # these are the only two we know of for now
    ex_dataset = ex_dataset.set_coords(
        [coord.full_nc_name for coord in COORDS if coord.full_nc_name in ex_dataset]
    )
    ex_dataset = sort_ds(ex_dataset)
    ex_dataset = set_axis_attrs(ex_dataset)
    ex_dataset = add_profile_type(ex_dataset, ftype=ftype)
    ex_dataset = add_geometry_var(ex_dataset)
    ex_dataset = finalize_ancillary_variables(ex_dataset)
    ex_dataset = combine_bottle_time(ex_dataset)
    ex_dataset = add_cdom_coordinate(ex_dataset)

    if _checks["flags"]:
        log.debug("Checking flags")
        check_flags(ex_dataset)

    return ex_dataset



import xarray as xr


class ExchangeError(ValueError):
    """This is the base exception which all the other exceptions derive from.
    It is a subclass of ValueError.
    """


# class ExchangeBottleError(ExchangeError):
#     pass
#
#
# class ExchangeCTDError(ExchangeError):
#     pass


class ExchangeEncodingError(ExchangeError):
    """Error raised when the bytes for some exchange file cannot be decoded as UTF-8."""


class ExchangeBOMError(ExchangeError):
    """Error raised when the exchange file has a byte order mark."""


class ExchangeLEError(ExchangeError):
    """Error raised when the exchange file does not have the correct line endings."""


class ExchangeMagicNumberError(ExchangeError):
    """Error raised when the exchange file does not start with ``BOTTLE`` or ``CTD``."""


class ExchangeEndDataError(ExchangeError):
    """Erorr raised when ``END_DATA`` cannot be found in the exchange file."""


class ExchangeParameterError(ExchangeError):
    """Base exception for errors related to parameters and units."""


class ExchangeParameterUndefError(ExchangeParameterError):
    """Error raised when the library does not have a definition for a parameter/unit pair in the exchange file."""

    def __init__(self, error_data: list[str]):
        super().__init__(error_data)
        self.error_data = error_data


class ExchangeParameterUnitAlignmentError(ExchangeParameterError):
    """Error raised when there is a mismatch between the number of parameters and number of units in the exchange file."""


class ExchangeDuplicateParameterError(ExchangeParameterError):
    """Error raised when the same parameter/unit pair occurs more than once in the excahnge file."""


class ExchangeOrphanFlagError(ExchangeParameterError):
    """Error raised when there exists a flag column with no corresponding parameter column."""


class ExchangeOrphanErrorError(ExchangeParameterError):
    """Error raised when there exists an error column with no corresponding parameter column."""


class ExchangeFlaglessParameterError(ExchangeParameterError):
    """Error raised when a parameter has a flag column when it is not supposed to."""


class ExchangeFlagUnitError(ExchangeParameterError):
    """Error raised if a flag column has a non empty units."""


class ExchangeDataError(ExchangeError):
    """Base exception for errors which occur when parsing the data porition of an exchange file."""


class ExchangeDataColumnAlignmentError(ExchangeDataError):
    """Error raised when the number of columns in a data line does not match the expected number of columns based on the parameter/unit lines."""


class ExchangeDataFlagPairError(ExchangeDataError):
    """There is a mismatch between what the flag value expects, and the fill/data value.

    Examples
    --------
    * something with a flag of ``9`` has a non fill value
    * something with a flag of ``2`` as a fill value instead of data
    """

    def __init__(self, error_data: xr.Dataset):
        super().__init__(error_data)
        self.error_data = error_data


class ExchangeDataPartialKeyError(ExchangeDataError):
    """Error raised when there is no value for one (or more) of the following parameters.

    * EXPOCODE
    * STNNBR
    * CASTNO
    * SAMPNO (only for bottle files)
    * CTDPRS (only for CTD files)

    These form the "composite key" which uniquely identify the "row" of exchange data.
    """


class ExchangeDuplicateKeyError(ExchangeDataError):
    """Error raised when there is a duplicate composite key in the exchange file.

    This would occur if the exact values for the following parameters occur in more than one data row:

    * EXPOCODE
    * STNNBR
    * CASTNO
    * SAMPNO (only for bottle files)
    * CTDPRS (only for CTD files)
    """


class ExchangeDataPartialCoordinateError(ExchangeDataError):
    """Error raised if values for latitude, longitude, or pressure are missing.

    It is OK by the standard to omit the time of day.
    """


class ExchangeDataInconsistentCoordinateError(ExchangeDataError):
    """Error raised if the reported latitude, longitude, and date (and time) vary for a single profile.

    A "profile" in an exchange file is a grouping of data rows which all have the same EXPOCODE, STNNBR, and CASTNO.
    The SAMPNO/CTDPRS is allowed/requried to vary for a single profile and is what identifies samples within one profile.
    """


class ExchangeInconsistentMergeType(ExchangeError):
    """Error raised when the merge_ex method is called on mixed ctd and bottle exchange types."""


class ExchangeRecursiveZip(ExchangeError):
    """Error raised if there are zip files inside the zip file that read exchange is trying to read."""



from importlib.resources import open_binary
from io import BytesIO

import numpy as np

from cchdo.hydro.exchange import read_csv


def test_read_csv():
    with open_binary("cchdo.hydro.tests.data", "btl_csv.csv") as f:
        test_data = f.read()

    read_csv(test_data)


def test_ignore_columns():
    test_data = BytesIO(
        b"""EXPOCODE,STNNBR,CASTNO,SAMPNO,LATITUDE,LONGITUDE,DATE,TIME,CTDPRS [DBAR],CTDOXY [UMOL/KG],CTDOXY [UMOL/KG]_FLAG_W,PROBLEM [UNIT],PROBLEM [UNIT]_FLAG_W
TEST,1,1,1,0,0,20220101,0000,0,0,2,0,3"""
    )
    read_csv(test_data, ignore_columns=("PROBLEM [UNIT]", "PROBLEM [UNIT]_FLAG_W"))


def test_all_flags_kept():
    test_data = BytesIO(
        b"""EXPOCODE,STNNBR,CASTNO,SAMPNO,LATITUDE,LONGITUDE,DATE,TIME,CTDPRS [DBAR],CTDOXY [UMOL/KG],CTDOXY [UMOL/KG]_FLAG_W,CTDOXY [UMOL/L],CTDOXY [UMOL/L]_FLAG_W
TEST,1,1,1,0,0,20220101,0000,0,0,2,0,3"""
    )
    ds = read_csv(test_data)

    assert "pressure" in ds
    assert "ctd_oxygen" in ds
    assert "ctd_oxygen_qc" in ds
    assert "ctd_oxygen_umol_l" in ds
    assert "ctd_oxygen_umol_l_qc" in ds
    assert ds.ctd_oxygen_qc.to_numpy() == [[2]]
    assert ds.ctd_oxygen_umol_l_qc.to_numpy() == [[3]]


def test_all_error_params():
    """Tests a condition where the presence of an error param was causing other params to be invalid (BTL DATE and TIME)

    Just needs to read without crashing
    """
    test_data = BytesIO(
        b"""EXPOCODE,STNNBR,CASTNO,SAMPNO,LATITUDE,LONGITUDE,DATE,TIME,CTDPRS [DBAR],TRITUM [KBQ/M^3],TRITUM [KBQ/M^3]_FLAG_W,TRITER [KBQ/M^3],BTL_DATE,BTL_TIME
TEST,1,1,1,0,0,20220101,0000,0,0,2,0,20220101,0000"""
    )
    ds = read_csv(test_data)

    assert "pressure" in ds


def test_station_ids():
    """Tests for a bug where station IDs were being truncated"""
    test_data = BytesIO(
        b"""EXPOCODE,SECT_ID,STNNBR,CASTNO,DATE,TIME,LATITUDE,LONGITUDE,CTDPRS [DBAR],CTDTMP [DEG C],CTDSAL [PSS-78],CTDOXY [UMOL/KG],CTDFLUOR [MG/M^3],CTDBEAMCP [/METER]
64PE20110724,AR07E,9,1,20110729,1919,59.57017,-38.77183,3006.0,1.2096,34.8913,304.0,0.0,0.161
64PE20110724,AR07E,10,1,20110730,105,59.46533,-37.77933,3.0,8.5907,34.6394,295.4,0.928,0.48
"""
    )
    ds = read_csv(test_data, ftype="C")
    np.testing.assert_equal(ds.station.values, ["9", "10"])



import xarray as xr

from cchdo.hydro.rename import rename_with_bookkeeping, to_argo_variable_names


def is_not_none(a):
    return a is not None


def test_rename():
    example = xr.Dataset(
        {
            "var0": xr.DataArray(
                [],
                attrs={
                    "standard_name": "cf_stdname1",
                    "ancillary_variables": "var1 var2",
                },
            ),
            "var1": xr.DataArray([], attrs={"standard_name": "cf_stdname2"}),
            "var2": xr.DataArray([], attrs={"standard_name": "cf_stdname3"}),
        }
    )
    name_dict = {"var1": "new_var1"}
    result = rename_with_bookkeeping(example, name_dict, attrs=["ancillary_variables"])
    assert "new_var1" in result.data_vars.keys()
    for da in result.filter_by_attrs(ancillary_variables=is_not_none).values():
        assert "new_var1" in da.attrs["ancillary_variables"]
        assert "var2" in da.attrs["ancillary_variables"]


def test_to_argo_variable_names():
    to_argo_variable_names("TODO")
